@@ -2,8 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2008  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2008-2009  Nokia Corporation
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -391,6 +391,40 @@ static int hold_call(struct csd_call *call)
 	return 0;
 }
 
+static int swap_calls(void)
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call(CSD_CALL_BUS_NAME, CSD_CALL_PATH,
+						CSD_CALL_INTERFACE,
+						"Swap");
+	if (!msg) {
+		error("Unable to allocate new D-Bus message");
+		return -ENOMEM;
+	}
+
+	g_dbus_send_message(connection, msg);
+
+	return 0;
+}
+
+static int create_conference(void)
+{
+	DBusMessage *msg;
+
+	msg = dbus_message_new_method_call(CSD_CALL_BUS_NAME, CSD_CALL_PATH,
+						CSD_CALL_INTERFACE,
+						"Conference");
+	if (!msg) {
+		error("Unable to allocate new D-Bus message");
+		return -ENOMEM;
+	}
+
+	g_dbus_send_message(connection, msg);
+
+	return 0;
+}
+
 static int call_transfer(void)
 {
 	DBusMessage *msg;
@@ -479,6 +513,9 @@ void telephony_answer_call_req(void *telephony_device)
 	if (!call)
 		call = find_call_with_status(CSD_CALL_STATUS_PROCEEDING);
 
+	if (!call)
+		call = find_call_with_status(CSD_CALL_STATUS_WAITING);
+
 	if (!call) {
 		telephony_answer_call_rsp(telephony_device,
 						CME_ERROR_NOT_ALLOWED);
@@ -557,9 +594,11 @@ void telephony_dial_number_req(void *telephony_device, const char *number)
 		number += 4;
 		flags = CALL_FLAG_PRESENTATION_RESTRICTED;
 	} else if (number[0] == '>') {
+		const char *location = &number[1];
+
 		number = memory_dial_lookup(strtol(&number[1], NULL, 0));
 		if (!number) {
-			error("No number at memory location %s", &number[1]);
+			error("No number at memory location %s", location);
 			telephony_dial_number_rsp(telephony_device,
 						CME_ERROR_INVALID_INDEX);
 			return;
@@ -719,6 +758,8 @@ void telephony_call_hold_req(void *telephony_device, const char *cmd)
 	switch (cmd[0]) {
 	case '0':
 		foreach_call_with_status(CSD_CALL_STATUS_HOLD, release_call);
+		foreach_call_with_status(CSD_CALL_STATUS_WAITING,
+								release_call);
 		break;
 	case '1':
 		if (idx) {
@@ -729,21 +770,35 @@ void telephony_call_hold_req(void *telephony_device, const char *cmd)
 		foreach_call_with_status(CSD_CALL_STATUS_ACTIVE, release_call);
 		call = find_call_with_status(CSD_CALL_STATUS_WAITING);
 		if (call)
-			answer_call(call);
+			err = answer_call(call);
 		break;
 	case '2':
 		if (idx) {
 			if (call)
 				err = split_call(call);
-			break;
+		} else {
+			struct csd_call *held, *wait;
+
+			call = find_call_with_status(CSD_CALL_STATUS_ACTIVE);
+			held = find_call_with_status(CSD_CALL_STATUS_HOLD);
+			wait = find_call_with_status(CSD_CALL_STATUS_WAITING);
+
+			if (wait)
+				err = answer_call(wait);
+			else if (call && held)
+				err = swap_calls();
+			else {
+				if (call)
+					err = hold_call(call);
+				if (held)
+					err = unhold_call(held);
+			}
 		}
-		foreach_call_with_status(CSD_CALL_STATUS_ACTIVE, hold_call);
-		foreach_call_with_status(CSD_CALL_STATUS_HOLD, unhold_call);
 		break;
 	case '3':
 		call = find_call_with_status(CSD_CALL_STATUS_HOLD);
 		if (call)
-			err = unhold_call(call);
+			err = create_conference();
 		break;
 	case '4':
 		err = call_transfer();
@@ -1518,7 +1573,7 @@ static void hal_find_device_reply(DBusPendingCall *call, void *user_data)
 {
 	DBusError err;
 	DBusMessage *reply;
-	DBusMessageIter iter, sub;;
+	DBusMessageIter iter, sub;
 	const char *path;
 	char match_string[256];
 	int type;
