@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -27,6 +27,7 @@ typedef void (*bt_io_callback_t) (GIOChannel *io, int err, const bdaddr_t *src,
 		const bdaddr_t *dst, gpointer user_data);
 typedef void (*bt_callback_t) (sdp_list_t *recs, int err, gpointer user_data);
 typedef void (*bt_destroy_t) (gpointer user_data);
+typedef void (*bt_hci_result_t) (uint8_t status, gpointer user_data);
 
 int bt_discover_services(const bdaddr_t *src, const bdaddr_t *dst,
 		bt_callback_t cb, void *user_data, bt_destroy_t destroy);
@@ -37,7 +38,8 @@ int bt_search_service(const bdaddr_t *src, const bdaddr_t *dst,
 int bt_cancel_discovery(const bdaddr_t *src, const bdaddr_t *dst);
 
 gchar *bt_uuid2string(uuid_t *uuid);
-uint16_t bt_string2class(const char *string);
+uint16_t bt_name2class(const char *string);
+char *bt_name2string(const char *string);
 int bt_string2uuid(uuid_t *uuid, const char *string);
 gchar *bt_list2string(GSList *list);
 GSList *bt_string2list(const gchar *str);
@@ -59,6 +61,9 @@ int bt_sco_connect(const bdaddr_t *src, const bdaddr_t *dst,
 GIOChannel *bt_sco_listen(const bdaddr_t *src, uint16_t mtu,
 				bt_io_callback_t cb, void *user_data);
 
+int bt_acl_encrypt(const bdaddr_t *src, const bdaddr_t *dst,
+			bt_hci_result_t cb, gpointer user_data);
+
 /* Experiemental bt_io API */
 
 typedef struct bt_io BtIO;
