@@ -4,7 +4,7 @@
  *
  *  Copyright (C) 2001-2002  Nokia Corporation
  *  Copyright (C) 2002-2003  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2002-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2002-2009  Marcel Holtmann <marcel@holtmann.org>
  *  Copyright (C) 2002-2003  Stephen Crane <steve.crane@rococosoft.com>
  *
  *
@@ -176,11 +176,13 @@ static int extract_des(uint8_t *buf, int len, sdp_list_t **svcReqSeq, uint8_t *p
 		case SDP_UUID128:
 			pElem = malloc(sizeof(uuid_t));
 			status = sdp_uuid_extract(p, bufsize, (uuid_t *) pElem, &localSeqLength);
-			if (status == 0) {
-				seqlen += localSeqLength;
-				p += localSeqLength;
-				bufsize -= localSeqLength;
+			if (status < 0) {
+				free(pElem);
+				return -1;
 			}
+			seqlen += localSeqLength;
+			p += localSeqLength;
+			bufsize -= localSeqLength;
 			break;
 		default:
 			return -1;
@@ -489,10 +491,11 @@ static int extract_attrs(sdp_record_t *rec, sdp_list_t *seq, uint8_t dtd, sdp_bu
 	if (!rec)
 		return SDP_INVALID_RECORD_HANDLE;
 
-	if (seq)
+	if (seq) {
 		SDPDBG("Entries in attr seq : %d", sdp_list_len(seq));
-	else
+	} else {
 		SDPDBG("NULL attribute descriptor");
+	}
 
 	SDPDBG("AttrDataType : %d", dtd);
 
@@ -826,7 +829,7 @@ done:
 	if (pattern)
 		sdp_list_free(pattern, free);
 	if (seq)
-                sdp_list_free(seq, free);
+		sdp_list_free(seq, free);
 	return status;
 }
 
