@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -480,7 +480,7 @@ static int avdtp_receive(int sk, void *data, int len)
 	return ret;
 }
 
-int avdtp_get_caps(int sk, int seid)
+static int avdtp_get_caps(int sk, int seid)
 {
 	struct seid_req req;
 	char buffer[1024];
@@ -511,7 +511,7 @@ int avdtp_get_caps(int sk, int seid)
 	return 0;
 }
 
-int avdtp_discover(int sk)
+static int avdtp_discover(int sk)
 {
 	struct avdtp_header req;
 	char buffer[256];
